#!/usr/bin/python
# vim: set fileencoding=iso-8859-2 :
#
# praca.py
# Bartomiej Brzozowiec, VIII 2007   (http://bb.homelinux.org/)
# Licencja: GPLv3 lub pniejsza. Zob. plik "GPLv3.txt" po szczegy.
#
# Wersje oprogramowania, z ktorym program byl pisany i testowany:
# - Qt - 3.3.7
# - python - 2.4.4
# - PyQt - 3.17  (sip - 4.5.2)
#
# - scipy - 0.5.2  (numpy - 1.0.1)
#

import sys
from qt import *
from scipy import *
from scipy.optimize import fsolve

def f2py( tabfstr ):
	"""
	Zamienia w stringach w tablicy tabfstr wszystkie wystapienia
		x1, x2, ..., xn  na
	        x[0], x[1], ..., x[n-1]
	w celu z-eval()-owania. Zwraca tabele ze zmienonymi lancuchami.
	"""
	newtabfstr = []
	for fstr in tabfstr:
		for j in range( 1, len( tabfstr ) + 1 ):
			fstr = fstr.replace( 'x' + str(j), 'x[' + str(j - 1) + ']' )
		newtabfstr.append( fstr )
	return newtabfstr

class myWidget( QWidget ):
	def __init__( self, parent = None, name = None ):
		QWidget.__init__( self, parent, name )
		
		self.setCaption( u'fsolve() - znajdywanie pierwiastkw ukadw rwna' )
		if name == None:
			self.setName( 'myWidget' )

		# glowny layout w pionie
		self.LayGlPion = QVBoxLayout( self, 6 )
		
		# layout na samej gorze, poziomy
		layGoraPoz = QHBoxLayout()
		self.LayGlPion.addLayout( layGoraPoz )
		
		opisGoraIle = QLabel( u'&Podaj ilo funkcji:', self )
		self.EdycjaGoraIle = QLineEdit( '2', self )
		self.N = None	# ilosc funkcji, int, ustawiana w funkcji okprzyc_f
		self.NZmienneStr = ''	# lancuch z nazwami dozwolonych zmiennych, tworzony w miejscu j.w.
		self.EdycjaGoraIle.setFocus()
		self.EdycjaGoraIle.setValidator( QIntValidator( 1, 15, self ) )
		# UWAGA: samo setValidator( QIntV...) nie zalatwia calkowicie sprawy - powinno sie sprawdzic,
		# czy nie jest puste albo czy nie jest zerem przed 'rzutowaniem' do int(..)...
		# (robimy to w okprzyc_f())
		opisGoraIle.setBuddy( self.EdycjaGoraIle )

		layGoraPoz.addWidget( opisGoraIle )
		layGoraPoz.addWidget( self.EdycjaGoraIle )

		self.connect( self.EdycjaGoraIle, SIGNAL( 'returnPressed()' ), self.okprzyc_f )
		
		self.PrzycOk = QPushButton( u'OK', self )
		self.connect( self.PrzycOk, SIGNAL( 'clicked()' ), self.okprzyc_f )
		layGoraPoz.addWidget( self.PrzycOk )

		przycKoniec = QPushButton( u'Koniec', self )
		przycKoniec.setFont( QFont( 'Times', 18, QFont.Bold ) )
		self.connect( przycKoniec, SIGNAL( 'clicked()' ), qApp, SLOT( 'quit()' ) )
		# mozna opuscic program naciskajac klawisz Escape:
		przycKoniec.setAccel( QKeySequence( 'Escape' ) )
		layGoraPoz.addWidget( przycKoniec )


		# kontenery na dynamicznie tworzone widgety
		self.LayoutyFunkcjePoz = []
		self.OpisyFunkcje = []
		self.EdycjeFunkcje = []

		self.LayoutyPunktStartPoz = []
		self.OpisyPunktStart = []
		self.EdycjePunktStart = []

		self.LayoutyJakobianPoz = []
		self.OpisyJakobian = []
		self.EdycjeJakobian = []

		self.PolaWyniku = []
		self.OpisyFodWynik = []
		self.PolaFodWynik = []
		self.InneWidgety = {}	# cztery labele, checkbox i trzy layouty, dwa przyciski i miejsce na komunikat od fsolve();
					# ilosc niezalezna od self.N

	# Funkcje
	# - zwiazane z dynamicznym tworzeniem i usuwaniem widgetow
	###
	def rob_widgety_funkcje( self ):
		#pass
		opisPodajFunkcje = QLabel( u'Podaj funkcje. Uyj ' + self.NZmienneStr + u' jako zmienne:', self )
		opisPodajFunkcje.setAlignment( Qt.AlignCenter )
		self.LayGlPion.addWidget( opisPodajFunkcje )
		opisPodajFunkcje.show()
		self.InneWidgety['opisPodajFunkcje'] = opisPodajFunkcje

		layFunkcjePoz = QHBoxLayout()
		self.LayGlPion.addLayout( layFunkcjePoz )
		self.LayoutyFunkcjePoz.append( layFunkcjePoz )
		for i in range( 1, self.N + 1 ):
			opisFn = QLabel( 'F' + str(i) + ':', self )	# np. "F1:"
			opisFn.show()
			layFunkcjePoz.addWidget( opisFn )
			self.OpisyFunkcje.append( opisFn )

			edycjaFn = QLineEdit( self, 'F' + str(i) )	# np. "F1"
			# TODO: odpowiedni walidator?
			edycjaFn.show()
			layFunkcjePoz.addWidget( edycjaFn )
			self.EdycjeFunkcje.append( edycjaFn )
	
	# usuwaj _w odwrotnej kolejnosci_ do dodawania!
	def usun_widgety_funkcje( self ):
		while self.EdycjeFunkcje:
			ed = self.EdycjeFunkcje.pop()
			ed.close( True )
		while self.OpisyFunkcje:
			op = self.OpisyFunkcje.pop()
			op.close( True )

		while self.LayoutyFunkcjePoz:
			lay = self.LayoutyFunkcjePoz.pop()
			self.LayGlPion.removeItem( lay )
		
		self.InneWidgety['opisPodajFunkcje'].close( True )
		del self.InneWidgety['opisPodajFunkcje']

	###
	def checkJakobian_f( self ):
		if self.InneWidgety['checkJakobian'].isChecked():
			# pokaz
			for op in self.OpisyJakobian:
				op.show()
			for ed in self.EdycjeJakobian:
				ed.show()
			self.EdycjeJakobian[0].setFocus()
		else:
			# ukryj
			for op in self.OpisyJakobian:
				op.hide()
			for ed in self.EdycjeJakobian:
				ed.hide()
	#
		
	def rob_widgety_punkt_startowy_i_checkbox( self ):
		opisPodajPunktStart = QLabel( u'Podaj punkt startowy:', self )
		opisPodajPunktStart.setAlignment( Qt.AlignCenter )
		self.LayGlPion.addWidget( opisPodajPunktStart )
		opisPodajPunktStart.show()
		self.InneWidgety['opisPodajPunktStart'] = opisPodajPunktStart

		layPunktStartPoz = QHBoxLayout()
		self.LayGlPion.addLayout( layPunktStartPoz )
		self.LayoutyPunktStartPoz.append( layPunktStartPoz )
		for i in range( 1, self.N + 1 ):
			opisPS = QLabel( 'x[' + str(i) + ']:', self )	# np. "x[1]:"
			opisPS.show()
			layPunktStartPoz.addWidget( opisPS )
			self.OpisyPunktStart.append( opisPS )

			edycjaPS = QLineEdit( '0.0', self, 'x' + str(i) )	# np. "x1"
			edycjaPS.setValidator( QDoubleValidator( self ) )
			edycjaPS.show()
			layPunktStartPoz.addWidget( edycjaPS )
			self.EdycjePunktStart.append( edycjaPS )

		# musi byc pojedynczy poziomy layout na pojedynczy widget (checkbox), zeby moc wysrodkowac
		# (QCheckBox nie ma metody setAlignment())
		layCheckJakobianPoz = QHBoxLayout()
		self.LayGlPion.addLayout( layCheckJakobianPoz )
		checkJakobian = QCheckBox( u'Podaj &jakobian (opcjonalnie)', self )
		#checkJakobian = QCheckBox( u'Podaj &jakobian (opcjonalnie).  Moesz uy ' + self.NZmienneStr + u' jako zmienne', self )
		checkJakobian.setChecked( False )
		#self.LayGlPion.addWidget( checkJakobian )
		layCheckJakobianPoz.addWidget( checkJakobian )
		layCheckJakobianPoz.setAlignment( Qt.AlignCenter )
		checkJakobian.show()
		self.connect( checkJakobian, SIGNAL( 'clicked()' ), self.checkJakobian_f )
		self.InneWidgety['checkJakobian'] = checkJakobian
		self.InneWidgety['layCheckJakobianPoz'] = layCheckJakobianPoz


	# usuwaj _w odwrotnej kolejnosci_ do dodawania!
	def usun_widgety_punkt_startowy_i_checkbox( self ):
		self.InneWidgety['checkJakobian'].close( True )
		del self.InneWidgety['checkJakobian']
		self.LayGlPion.removeItem( self.InneWidgety['layCheckJakobianPoz'] )
		del self.InneWidgety['layCheckJakobianPoz']

		while self.EdycjePunktStart:
			ed = self.EdycjePunktStart.pop()
			ed.close( True )
		while self.OpisyPunktStart:
			op = self.OpisyPunktStart.pop()
			op.close( True )

		while self.LayoutyPunktStartPoz:
			lay = self.LayoutyPunktStartPoz.pop()
			self.LayGlPion.removeItem( lay )

		self.InneWidgety['opisPodajPunktStart'].close( True )
		del self.InneWidgety['opisPodajPunktStart']

	###
	# tworzy widgety, ale domyslnie nie pokazuje (bo domyslnie checkbox nie jest zaznaczony)
	def rob_widgety_jakobian( self ):
		idx = 0
		for i in range( 1, self.N + 1 ):
			layJakobianPoz = QHBoxLayout()
			self.LayGlPion.addLayout( layJakobianPoz )
			self.LayoutyJakobianPoz.append( layJakobianPoz )
			for j in range( 1, self.N + 1 ):
				idx += 1
				opisJ = QLabel( 'dF' + str(i) + '/dx' + str(j) + ':', self )	# np. "dF1/dx1:"
				#opisJ.show()	# nie pokazujemy, dopiero po zaznaczeniu checkboxa
				layJakobianPoz.addWidget( opisJ )
				self.OpisyJakobian.append( opisJ )

				edycjaJ = QLineEdit( self, 'F' + str(i) + 'x' + str(j) + '_' + str(idx) )	# np. "F1x1_1"
				# TODO: odpowiedni walidator?
				#edycjaJ.show()	# nie pokazujemy, dopiero po zaznaczeniu checkboxa
				layJakobianPoz.addWidget( edycjaJ )
				self.EdycjeJakobian.append( edycjaJ )
				

	# usuwaj _w odwrotnej kolejnosci_ do dodawania!
	def usun_widgety_jakobian( self ):
		while self.EdycjeJakobian:
			ed = self.EdycjeJakobian.pop()
			ed.close( True )
		while self.OpisyJakobian:
			op = self.OpisyJakobian.pop()
			op.close( True )

		while self.LayoutyJakobianPoz:
			lay = self.LayoutyJakobianPoz.pop()
			self.LayGlPion.removeItem( lay )

	###
	# przycisk "Rozwi!", label "Wynik:" i "X0:", N pl na wynik, N pl na wartosc funkcji w wyniku z N labelami,
	# pole na komunikat od fsolve(), przycisk "Wprowad inn ilo funkcji"
	def rob_widgety_dol( self ):
		przycRozwiaz = QPushButton( u'&Rozwi!', self )
		self.connect( przycRozwiaz, SIGNAL( 'clicked()' ), self.rozwiaz_f )
		self.LayGlPion.addWidget( przycRozwiaz )
		przycRozwiaz.show()
		self.InneWidgety['przycRozwiaz'] = przycRozwiaz

		opisWynik = QLabel( u'Wynik:', self )
		opisWynik.setAlignment( Qt.AlignCenter )
		self.LayGlPion.addWidget( opisWynik )
		self.InneWidgety['opisWynik'] = opisWynik

		layWynik = QHBoxLayout()
		self.LayGlPion.addLayout( layWynik )
		self.InneWidgety['layWynik'] = layWynik
		opisX0 = QLabel( u'X0:', self )
		layWynik.addWidget( opisX0 )
		self.InneWidgety['opisX0'] = opisX0
		for i in range( 1, self.N + 1 ):
			# LineEdit, ale "read-only" tak naprawde - tylko do pokazywania wyniku
			edycjaWynikRO = QLineEdit( self, 'wynik' + str(i) )	# np. "wynik1"
			edycjaWynikRO.setReadOnly( True )	# zasymuluj label z mozliwoscia zaznaczania
								# (Hint: w nowszym Qt (>=4.2?) QLabel ma mozliwosc ustawienia
								# opcji, ze mozna jego zawartosc zaznaczac...
								# .setTextInteractionFlags( Qt.TextSelectableByMouse ) )
			edycjaWynikRO.setFrameShadow( QFrame.Sunken )	# j.w., c.d.
			edycjaWynikRO.setFrameShape( QFrame.MenuBarPanel )	# j.w., c.d.
			edycjaWynikRO.setPaletteBackgroundColor( QColor( 'lightgray' ) )	# j.w., c.d.
			layWynik.addWidget( edycjaWynikRO )
			self.PolaWyniku.append( edycjaWynikRO )

		layFodWynik = QHBoxLayout()
		self.LayGlPion.addLayout( layFodWynik )
		self.InneWidgety['layFodWynik'] = layFodWynik
		for i in range( 1, self.N + 1 ):
			opisFodWynik = QLabel( 'F' + str(i) + '(X0):', self )	# np. "F1(X0):"
			layFodWynik.addWidget( opisFodWynik )
			self.OpisyFodWynik.append( opisFodWynik )

			# LineEdit, ale "read-only" tak naprawde - tylko do pokazywania wartosci funkcji w wyniku
			edycjaFodWynikRO = QLineEdit( self, 'Fodwynik' + str(i) )	# np. "Fodwynik1"
			edycjaFodWynikRO.setReadOnly( True )	# zasymuluj label z mozliwoscia zaznaczania
								# (Hint: w nowszym Qt (>=4.2?) QLabel ma mozliwosc ustawienia
								# opcji, ze mozna jego zawartosc zaznaczac...
								# .setTextInteractionFlags( Qt.TextSelectableByMouse ) )
			edycjaFodWynikRO.setFrameShadow( QFrame.Sunken )	# j.w., c.d.
			edycjaFodWynikRO.setFrameShape( QFrame.MenuBarPanel )	# j.w., c.d.
			edycjaFodWynikRO.setPaletteBackgroundColor( QColor( 'lightgray' ) )	# j.w., c.d.
			layFodWynik.addWidget( edycjaFodWynikRO )
			self.PolaFodWynik.append( edycjaFodWynikRO )

		# komunikat od fsolve()
		edycjaFSolveMessRO = QLineEdit( self )
		edycjaFSolveMessRO.setReadOnly( True )	# patrz wyej
		edycjaFSolveMessRO.setFrameShadow( QFrame.Sunken )
		edycjaFSolveMessRO.setFrameShape( QFrame.MenuBarPanel )
		edycjaFSolveMessRO.setPaletteBackgroundColor( QColor( 'lightgray' ) )
		self.LayGlPion.addWidget( edycjaFSolveMessRO )
		self.InneWidgety['edycjaFSolveMessRO'] = edycjaFSolveMessRO

		przycWprowadzInna = QPushButton( u'&Wprowad inn ilo funkcji', self )
		self.connect( przycWprowadzInna, SIGNAL( 'clicked()' ), self.wprowadz_inna_f )
		self.LayGlPion.addWidget( przycWprowadzInna )
		przycWprowadzInna.show()
		self.InneWidgety['przycWprowadzInna'] = przycWprowadzInna
	
	# usuwaj _w odwrotnej kolejnosci_ do dodawania!
	def usun_widgety_dol( self ):
		self.InneWidgety['przycWprowadzInna'].close( True )
		del self.InneWidgety['przycWprowadzInna']

		self.InneWidgety['edycjaFSolveMessRO'].close( True )
		del self.InneWidgety['edycjaFSolveMessRO']

		while self.PolaFodWynik:
			fodwyn = self.PolaFodWynik.pop()
			fodwyn.close( True )
		while self.OpisyFodWynik:
			op = self.OpisyFodWynik.pop()
			op.close( True )
		
		self.LayGlPion.removeItem( self.InneWidgety['layFodWynik'] )
		del self.InneWidgety['layFodWynik']
		
		while self.PolaWyniku:
			wyn = self.PolaWyniku.pop()
			wyn.close( True )

		self.InneWidgety['opisX0'].close( True )
		del self.InneWidgety['opisX0']
		
		self.LayGlPion.removeItem( self.InneWidgety['layWynik'] )
		del self.InneWidgety['layWynik']

		self.InneWidgety['opisWynik'].close( True )
		del self.InneWidgety['opisWynik']

		self.InneWidgety['przycRozwiaz'].close( True )
		del self.InneWidgety['przycRozwiaz']
	
	# ----

	# metoda wywolywana po nacisnieciu przycisku "Rozwi!"
	def rozwiaz_f( self ):
		# TODO(?): walidacja wprowadzonych danych, bardziej zaawansowana niz tylko sprawdzanie, czy nie ma pustych...
		# funkcje
		F = [str(f.text()) for f in self.EdycjeFunkcje]
		if '' in F:
			QMessageBox.warning( self, u'Bd', u'Prosz wypeni wszystkie pola funkcji', QMessageBox.Ok, QMessageBox.NoButton )
			# ustawia fokus zawsze na pierwszym... (TODO...)
			self.EdycjeFunkcje[0].setFocus()
			return
		
		# jakobian (jesli uzytkownik chcial go podac)
		Fjak = [str(fjak.text()) for fjak in self.EdycjeJakobian]
		if self.InneWidgety['checkJakobian'].isChecked():
			if '' in Fjak:
				QMessageBox.warning( self, u'Bd', u'Prosz wypeni wszystkie pola jakobianu', QMessageBox.Ok, QMessageBox.NoButton )
				self.EdycjeJakobian[0].setFocus()
				return

		# punkt startowy
		try:
			Fsp = [float(sp.text()) for sp in self.EdycjePunktStart]
		except ValueError:
			QMessageBox.warning( self, u'Bd', u'Prosz poda punkt startowy', QMessageBox.Ok, QMessageBox.NoButton )
			self.EdycjePunktStart[0].setFocus()
			return

		# konwersja do formatu, ktory python moze z-eval()-owac
		F_py = f2py( F )
		Fjak_py = f2py( Fjak )

		# definicje lokalnych funkcji dla fsolve()
		# (nazwa zmiennej ("x") koresponduje z nazwa zmiennej uzytej przy podawaniu funkcji ("x1", "x2" itd.)!)
		def func( x ):
			out = []
			for f in F_py:
				out.append( eval( f ) )
			return out
		
		N = self.N
		def func_jak( x ):
			out = []
			row = []
			for i, fprim in enumerate( Fjak_py ):
				row.append( eval( fprim ) )
				if (i+1) % N == 0:
					out.append( row )
					row = []
			return out

		# glowny punkt programu - wywolanie fsolve() ze SciPy
		if self.InneWidgety['checkJakobian'].isChecked():
			x0, infodict, ier, mesg = fsolve( func, Fsp, fprime=func_jak, full_output=1 )
		else:
			x0, infodict, ier, mesg = fsolve( func, Fsp, full_output=1 )

		# pokaz wynik
		self.pokaz_widgety_wynik()
		# zalatw przypadek pojedynczej zmiennej (czyli dla self.N=1):
		if type(x0) != type(()) and type(x0) != type([]) and type(x0) != type(array(())):
			x0 = (x0,)
		if ier == 1:	# fsolve() zakonczylo dzialanie pomyslnie, tzn. znaleziono rozwiazanie
			for i, x in enumerate( x0 ):
				self.PolaWyniku[i].setText( str(x) )
			for i, fodx in enumerate( infodict['fvec'] ):
				self.PolaFodWynik[i].setText( str(fodx) )
		else:
			for wyn in self.PolaWyniku:
				wyn.setText( '##' )
			for fodwyn in self.PolaFodWynik:
				fodwyn.setText( '##' )
		self.InneWidgety['edycjaFSolveMessRO'].setText( mesg )

	
	# pokazuje widgety zwiazane z prezentacja wyniku dzialania fsolve(); czysci tez pola...
	def pokaz_widgety_wynik( self ):
		self.InneWidgety['opisWynik'].show()
		self.InneWidgety['opisX0'].show()
		for wyn in self.PolaWyniku:
			wyn.show()
			wyn.setText( '' )
		for op in self.OpisyFodWynik:
			op.show()
		for fodwyn in self.PolaFodWynik:
			fodwyn.show()
			fodwyn.setText( '' )
		self.InneWidgety['edycjaFSolveMessRO'].show()
		self.InneWidgety['edycjaFSolveMessRO'].setText( '' )
	
	# metoda wywolywana po nacisnieciu przycisku "Wprowad inn ilo funkcji"
	def wprowadz_inna_f( self ):
		self.usun_widgety_dol()
		self.usun_widgety_jakobian()
		self.usun_widgety_punkt_startowy_i_checkbox()
		self.usun_widgety_funkcje()

		self.EdycjaGoraIle.setFocus()
		self.PrzycOk.setEnabled( True )	# przycisk zostal zdeaktywowany w okprzyc_f() wczesniej...

	# metoda wywolywana po nacisnieciu Enter w polu edycji ilosci funkcji albo po wybraniu przycisku "OK"
	def okprzyc_f( self ):
		if not self.PrzycOk.isEnabled(): return
		if self.EdycjaGoraIle.text() != '':
			self.N = int( self.EdycjaGoraIle.text() )
			if self.N <= 0:
				return
		else:
			return
		self.PrzycOk.setEnabled( False )	# nie zapomnij pozniej wlaczyc!

		assert self.N != None

		zmStrTab = []
		for i in range( 1, self.N + 1 ):
			zmStrTab.append( '"x' + str( i ) + '"' )
		self.NZmienneStr = ', '.join( zmStrTab )

		self.rob_widgety_funkcje()
		self.EdycjeFunkcje[0].setFocus()
		self.rob_widgety_punkt_startowy_i_checkbox()
		self.rob_widgety_jakobian()	# zawsze rob, ale domyslnie od razu nie pokazuj
		self.rob_widgety_dol()
		# usuwaj najlepiej _w odwrotnej kolejnosci_ do dodawania! (usuwamy w wprowadz_inna_f())


# punkt wejscia do programu:
if __name__ == '__main__':
	app = QApplication( sys.argv )
	QObject.connect( app, SIGNAL( 'lastWindowClosed()' ),
			 app, SLOT( 'quit()' ) )
	win = myWidget()
	win.setGeometry( QRect( 190, 218, 478, 108 ) )
	app.setMainWidget( win )
	win.show()
	app.exec_loop()
